"use strict";

var type = require("type-detect");
var clone = require("clone");
var associative = require("associative");

(function () {
    Array.prototype.forAll = Array.prototype.forAll || function (func) {
        for (var i = 0; i < this.length; i++) {
            if (func) func(this[i], i);
        }
    };

    Array.prototype.select = Array.prototype.select || function (func) {
        var arr = [];
        for (var i = 0; i < this.length; i++) {
            arr.push(func ? func(this[i], i) : this[i]);
        }return arr;
    };

    Array.prototype.distinct = Array.prototype.distinct || function () {
        var u = {},
            a = [];
        for (var i = 0, l = this.length; i < l; ++i) {
            if (u.hasOwnProperty(this[i])) continue;
            a.push(this[i]);
            u[this[i]] = 1;
        }
        return a;
    };

    Array.prototype.where = Array.prototype.where || function (func) {
        var arr = [];
        for (var i = 0; i < this.length; i++) {
            if (!func || func(this[i], i)) arr.push(this[i]);
        }return arr;
    };

    Array.prototype.firstOrDefault = Array.prototype.firstOrDefault || function (predicate) {
        if (!predicate) return this.length > 0 ? this[0] : null;

        for (var i = 0; i < this.length; i++) {
            if (predicate(this[i]) === true) return this[i];
        }return null;
    };

    Array.prototype.first = Array.prototype.first || function (predicate) {
        var firstOrDefault = this.firstOrDefault(predicate);
        if (firstOrDefault === null) throw new Error("First item not found");else return firstOrDefault;
    };

    Array.prototype.lastOrDefault = Array.prototype.lastOrDefault || function (predicate) {
        if (!predicate) return this.length > 0 ? this[this.length - 1] : null;

        for (var i = this.length - 1; i >= 0; i--) {
            if (predicate(this[i]) === true) return this[i];
        }return null;
    };

    Array.prototype.last = Array.prototype.last || function (predicate) {
        var lastOrDefault = this.lastOrDefault(predicate);
        if (lastOrDefault === null) throw new Error("Last item not found");else return lastOrDefault;
    };

    Array.prototype.all = Array.prototype.all || function (func) {
        return this.where(func).length === this.length;
    };

    Array.prototype.any = Array.prototype.any || function (func) {
        return this.firstOrDefault(func) !== null;
    };

    Array.prototype.accumulate = Array.prototype.accumulate || function (initial, func) {
        // initial parameter is required
        var accum = initial;
        for (var i = 0; i < this.length; i++) {
            accum = func ? func(accum, this[i]) : accum + this[i];
        }return accum;
    };

    Array.prototype.take = Array.prototype.take || function (count) {
        var arr = [];
        for (var i = 0; i < count; i++) {
            arr.push(this[i]);
        }return arr;
    };

    Array.prototype.skip = Array.prototype.skip || function (count) {
        var arr = [];
        for (var i = count; i < this.length; i++) {
            arr.push(this[i]);
        }return arr;
    };

    Array.prototype.union = Array.prototype.union || function (array) {
        var arr = [];

        for (var i = 0; i < this.length; i++) {
            arr.push(this[i]);
        }if (type(array) === "array") for (var k = 0; k < array.length; k++) {
            arr.push(array[k]);
        }return arr;
    };

    Array.prototype.orderBy = Array.prototype.orderBy || function (func) {
        return this.select().sort(function (a, b) {
            var objA = func ? func(a) : a;
            var objB = func ? func(b) : b;

            var comparison = null;
            if (objA && objB) {
                if (type(objA) === "number") comparison = objA - objB;else if (type(objA) === "date") comparison = objA.getTime() - objB.getTime();else comparison = objA.toString().localeCompare(objB.toString());
            } else {
                if (!objA && !objB) comparison = 0;else if (!objA) comparison = -1;else comparison = 1;
            }
            return comparison;
        });
    };

    Array.prototype.orderByDesc = Array.prototype.orderByDesc || function (func) {
        return this.orderBy(func).reverse();
    };

    Array.prototype.sum = Array.prototype.sum || function (func) {
        var sum = 0;
        for (var i = 0; i < this.length; i++) {
            sum += func ? func(this[i]) : 0;
        }return sum;
    };

    Array.prototype.count = Array.prototype.count || function (func) {
        var count = 0;
        for (var i = 0; i < this.length; i++) {
            if (true === (func ? func(this[i]) : this[i])) count++;
        }return count;
    };

    Array.prototype.avg = Array.prototype.avg || function (func) {
        var count = this.count(func);
        if (count === 0) return 0;else return this.sum(func) / count;
    };

    Array.prototype.max = Array.prototype.max || function (func) {
        var max = Number.NEGATIVE_INFINITY;
        for (var i = 0; i < this.length; i++) {
            var value = func ? func(this[i]) : this[i];
            if (value > max) max = value;
        }
        return max;
    };

    Array.prototype.min = Array.prototype.min || function (func) {
        var min = Number.POSITIVE_INFINITY;
        for (var i = 0; i < this.length; i++) {
            var value = func ? func(this[i]) : this[i];
            if (value < max) min = value;
        }
        return min;
    };

    // Groups the array values into a dictionary of arrays. 
    // The grouping keys are determined by applying the <func> function argument to each element 
    Array.prototype.groupBy = Array.prototype.groupBy || function (func) {
        var groups = new associative.Dictionary();
        for (var i = 0; i < this.length; i++) {
            var key = func(this[i]);
            if (!groups.hasKey(key)) groups.put(key, []);
            groups.get(key).push(this[i]);
        }
        return groups;
    };

    // Groups this array into an dictionary of <keyCount> arrays,
    // each containing the same number of elements, except possibly the last array, 
    // which may have less elements than the others 
    // if the length of this array is not divisible by <keyCount>
    Array.prototype.groupByNumber = Array.prototype.groupByNumber || function (keyCount) {
        var groups = new associative.Dictionary();
        var groupLength = Math.floor(this.length / keyCount);
        var remainder = this.length % count;

        for (var i = 0; i < keyCount; i++) {
            if (!groups.hasKey(i)) groups.put(i, []);

            groups.get(i).append(this.slice(i * groupLength, (i + 1) * groupLength));
        }
        return groups;
    };

    // groups this array according to <key> generator function and then 
    // aggregates the resulting dictionary over <value> generator function
    Array.prototype.groupAggregate = Array.prototype.groupAggregate || function (key, value, aggregation) {
        // if key is passed as generator function, use it, else create generator function that assumes key is a property name. 
        // If key is null/missing, the string representation of an object is used as the key. 
        var keyGenerator = type(key) === "function" ? key : function (datum) {
            return key ? datum[key.toString()] : datum.toString();
        };

        //Same for value
        var valueGenerator = type(value) === "function" ? value : function (datum) {
            return value ? datum[value.toString()] : datum.toString();
        };

        return this.groupBy(keyGenerator).select(function (groupData) {
            return (// e.g., groupData.sum(valueGenerator)
                groupData[aggregation](valueGenerator)
            );
        });
    };

    Array.prototype.removeItem = Array.prototype.removeItem || function (item) {
        var _index = this.indexOf(item);
        if (_index < 0) throw "removeItem(): item not found in array";
        return this.splice(_index, 1, 0);
    };

    Array.prototype.removeAt = Array.prototype.removeAt || function (index, count) {
        return this.splice(index, count, 0);
    };

    Array.prototype.insert = Array.prototype.insert || function (index) {
        for (var _len = arguments.length, items = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
            items[_key - 1] = arguments[_key];
        }

        return this.splice(index, 0, items);
    };

    Array.prototype.append = Array.prototype.append || function (index) {
        return this.splice(this.length, 0, items);
    };

    Array.prototype.flatten = Array.prototype.flatten || function () {
        var mutable = false;
        var array = this;

        var toString = Object.prototype.toString;
        var arrayTypeStr = '[object Array]';

        var result = [];
        var nodes = mutable && array || array.slice();
        var node;

        if (!array.length) {
            return result;
        }

        node = nodes.pop();

        do {
            if (toString.call(node) === arrayTypeStr) {
                nodes.push.apply(nodes, node);
            } else {
                result.push(node);
            }
        } while (nodes.length && (node = nodes.pop()) !== undefined);

        result.reverse(); // we reverse result to restore the original order
        return result;
    };
})();

/*module.exports.testInfo = {
    getContext: () => {
        module.exports.extendArray();
    },
    defaultUnit: function (method, ...args) {
        let arr = [
            new Date(),
            /abc[.]+/g,
            "a", "b", "c",
            1, 2, 3, 4, 5, 6, 7,
            [
                0,
                1
            ],
            [
                2,
                3
            ],
            [
                4,
                5,
                [
                    6,
                    7,
                    [
                        8,
                        [
                            9,
                            10
                        ]
                    ]
                ]
            ]
        ];

        return Array.prototype[method].call(arr, ...args);
    },
    cases: [
        {
            title: "where() method returns correct array",
            getArgs: (ctx) => ["where", ((x) => type(x) === "number" && x > 3)],
            getExpected: (ctx) => [4, 5, 6, 7]
        }
    ]
};
*/